@@ -1,5 +1,5 @@
 /*********************************************************
- * Copyright (C) 2003-2015 VMware, Inc. All rights reserved.
+ * Copyright (c) 2003-2023 VMware, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -49,6 +49,13 @@
 #define INCLUDE_ALLOW_VMCORE
 #include "includeCheck.h"
 
+#if defined _MSC_VER && !defined BORA_NO_WIN32_INTRINS
+#pragma warning(push)
+#pragma warning(disable : 4255)      // disable no-prototype() to (void) warning
+#include <intrin.h>
+#pragma warning(pop)
+#endif
+
 #include "vm_basic_types.h"
 
 #if defined VM_X86_64
@@ -57,14 +64,28 @@
 #elif defined VM_X86_32
 #include "vm_basic_asm_x86_common.h"
 #include "vm_basic_asm_x86.h"
+#elif defined VM_ARM_32
+#include "vm_basic_asm_arm32.h"
+#define MUL64_NO_ASM 1
+#include "mul64.h"
 #elif defined VM_ARM_64
-#include "arm64_basic_defs.h"
 #include "vm_basic_asm_arm64.h"
 #else
 #define MUL64_NO_ASM 1
 #include "mul64.h"
 #endif
 
+#if defined _M_ARM64EC || defined _M_ARM64
+#include "vm_assert.h"
+#define MUL64_NO_ASM 1
+#include "mul64.h"
+#endif
+
+#if defined __cplusplus
+extern "C" {
+#endif
+
+
 /*
  * Locate most and least significant bit set functions. Use our own name
  * space to avoid namespace collisions. The new names follow a pattern,
@@ -94,7 +115,7 @@
  * mssb64      MSB set (uint64)            1..64    0
  */
 
-#ifdef _MSC_VER
+#if defined(_MSC_VER) && !defined(__clang__) // Clang defines _MSC_VER on Windows
 static INLINE int
 lssb32_0(const uint32 value)
 {
@@ -109,7 +130,7 @@ lssb32_0(const uint32 value)
    __analysis_assume(ret != 0);
 #endif
 
-#pragma warning(suppress: 6102) // Suppress bogus complaint that idx may be uninitialized in error case
+#pragma warning(suppress: 6001 6102) // Suppress bogus complaint that idx may be uninitialized in error case
    return idx;
 }
 
@@ -127,7 +148,7 @@ mssb32_0(const uint32 value)
    __analysis_assume(ret != 0);
 #endif
 
-#pragma warning(suppress: 6102) // Suppress bogus complaint that idx may be uninitialized in error case
+#pragma warning(suppress: 6001 6102) // Suppress bogus complaint that idx may be uninitialized in error case
    return idx;
 }
 
@@ -146,7 +167,7 @@ lssb64_0(const uint64 value)
       __analysis_assume(ret != 0);
 #endif
 
-#pragma warning(suppress: 6102) // Suppress bogus complaint that idx may be uninitialized in error case
+#pragma warning(suppress: 6001 6102) // Suppress bogus complaint that idx may be uninitialized in error case
       return idx;
 #else
       /* The coding was chosen to minimize conditionals and operations */
@@ -177,7 +198,7 @@ mssb64_0(const uint64 value)
       __analysis_assume(ret != 0);
 #endif
 
-#pragma warning(suppress: 6102) // Suppress bogus complaint that idx may be uninitialized in error case
+#pragma warning(suppress: 6001 6102) // Suppress bogus complaint that idx may be uninitialized in error case
       return idx;
 #else
       /* The coding was chosen to minimize conditionals and operations */
@@ -190,7 +211,7 @@ mssb64_0(const uint64 value)
 }
 #endif
 
-#ifdef __GNUC__
+#if defined __GNUC__ || defined __clang__
 
 #ifdef VM_X86_ANY
 #define USE_ARCH_X86_CUSTOM
@@ -221,13 +242,10 @@ mssb64_0(const uint64 value)
  * **********************************************************
  */
 
-#if __GNUC__ < 4
-#define FEWER_BUILTINS
-#endif
-
 static INLINE int
-lssb32_0(uint32 value)
+lssb32_0(uint32 v)
 {
+   int value = (int)v;
 #ifdef USE_ARCH_X86_CUSTOM
    if (!__builtin_constant_p(value)) {
       if (UNLIKELY(value == 0)) {
@@ -242,7 +260,6 @@ lssb32_0(uint32 value)
    return __builtin_ffs(value) - 1;
 }
 
-#ifndef FEWER_BUILTINS
 static INLINE int
 mssb32_0(uint32 value)
 {
@@ -267,8 +284,10 @@ mssb32_0(uint32 value)
 }
 
 static INLINE int
-lssb64_0(const uint64 value)
+lssb64_0(const uint64 v)
 {
+   int64 value = (int64)v;
+
 #ifdef USE_ARCH_X86_CUSTOM
    if (!__builtin_constant_p(value)) {
       if (UNLIKELY(value == 0)) {
@@ -293,46 +312,6 @@ lssb64_0(const uint64 value)
 #endif
    return __builtin_ffsll(value) - 1;
 }
-#endif /* !FEWER_BUILTINS */
-
-#ifdef FEWER_BUILTINS
-/* GCC 3.3.x does not like __bulitin_clz or __builtin_ffsll. */
-static INLINE int
-mssb32_0(uint32 value)
-{
-   if (UNLIKELY(value == 0)) {
-      return -1;
-   } else {
-      int pos;
-      __asm__ __volatile__("bsrl %1, %0\n" : "=r" (pos) : "rm" (value) : "cc");
-      return pos;
-   }
-}
-
-static INLINE int
-lssb64_0(const uint64 value)
-{
-   if (UNLIKELY(value == 0)) {
-      return -1;
-   } else {
-      intptr_t pos;
-
-#ifdef VM_X86_64
-      __asm__ __volatile__("bsf %1, %0\n" : "=r" (pos) : "rm" (value) : "cc");
-#else
-      /* The coding was chosen to minimize conditionals and operations */
-      pos = lssb32_0((uint32) value);
-      if (pos == -1) {
-         pos = lssb32_0((uint32) (value >> 32));
-         if (pos != -1) {
-            return pos + 32;
-         }
-      }
-#endif /* VM_X86_64 */
-      return pos;
-   }
-}
-#endif /* FEWER_BUILTINS */
 
 
 static INLINE int
@@ -378,10 +357,10 @@ lssbPtr_0(const uintptr_t value)
 #endif
 }
 
-static INLINE int
+static INLINE unsigned
 lssbPtr(const uintptr_t value)
 {
-   return lssbPtr_0(value) + 1;
+   return (unsigned)lssbPtr_0(value) + 1;
 }
 
 static INLINE int
@@ -394,41 +373,39 @@ mssbPtr_0(const uintptr_t value)
 #endif
 }
 
-static INLINE int
+static INLINE unsigned
 mssbPtr(const uintptr_t value)
 {
-   return mssbPtr_0(value) + 1;
+   return (unsigned)mssbPtr_0(value) + 1;
 }
 
-static INLINE int
+static INLINE unsigned
 lssb32(const uint32 value)
 {
-   return lssb32_0(value) + 1;
+   return (unsigned)lssb32_0(value) + 1;
 }
 
-static INLINE int
+static INLINE unsigned
 mssb32(const uint32 value)
 {
-   return mssb32_0(value) + 1;
+   return (unsigned)mssb32_0(value) + 1;
 }
 
-static INLINE int
+static INLINE unsigned
 lssb64(const uint64 value)
 {
-   return lssb64_0(value) + 1;
+   return (unsigned)lssb64_0(value) + 1;
 }
 
-static INLINE int
+static INLINE unsigned
 mssb64(const uint64 value)
 {
-   return mssb64_0(value) + 1;
+   return (unsigned)mssb64_0(value) + 1;
 }
 
-#ifdef __GNUC__
-#if defined(VM_X86_ANY) || defined(VM_ARM_ANY)
 
 /*
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  *
  * uint16set --
  *
@@ -440,16 +417,18 @@ mssb64(const uint64 value)
  * Side effects:
  *      As with memset.
  *
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  */
 
-static INLINE void *
-uint16set(void *dst, uint16 val, size_t count)
+static inline void *
+uint16set(void *dst,    // OUT
+          uint16 val,   // IN
+          size_t count) // IN
 {
-#ifdef VM_ARM_32
+#if defined __GNUC__ && defined VM_ARM_32
    void *tmpDst = dst;
 
-   __asm__ __volatile__ (
+   __asm__ __volatile__(
       "cmp     %1, #0\n\t"
       "beq     2f\n"
       "1:\n\t"
@@ -460,7 +439,7 @@ uint16set(void *dst, uint16 val, size_t count)
       : "+r" (tmpDst), "+r" (count)
       : "r" (val)
       : "cc", "memory");
-#elif defined(VM_ARM_64)
+#elif defined __GNUC__ && defined VM_ARM_64
    void   *tmpDst = dst;
    uint64  tmpVal = 0;
 
@@ -468,7 +447,7 @@ uint16set(void *dst, uint16 val, size_t count)
       return dst;
    }
 
-   __asm__ __volatile__ (
+   __asm__ __volatile__(
       "cbz     %3, 1f\n\t"
 
       // Copy 16 bits twice...
@@ -502,26 +481,33 @@ uint16set(void *dst, uint16 val, size_t count)
       "strh    %w2, [%0]\n"
       "4:"
       : "+r" (tmpDst), "+r" (count), "+r" (tmpVal)
-      : "r" (val)
+      : "r" ((uint64)val)
       : "cc", "memory");
-#else
+#elif defined __GNUC__ && defined VM_X86_ANY
    size_t dummy0;
    void *dummy1;
 
-   __asm__ __volatile__("\t"
-                        "cld"            "\n\t"
-                        "rep ; stosw"    "\n"
-                        : "=c" (dummy0), "=D" (dummy1)
-                        : "0" (count), "1" (dst), "a" (val)
-                        : "memory", "cc"
-      );
+   __asm__ __volatile__(
+      "cld"            "\n\t"
+      "rep ; stosw"    "\n"
+      : "=c" (dummy0), "=D" (dummy1)
+      : "0" (count), "1" (dst), "a" (val)
+      : "memory", "cc");
+#elif defined _MSC_VER && defined VM_X86_ANY
+   __stosw((uint16 *)dst, val, count);
+#else /* Fall back to a plain C implementation. */
+   size_t i;
+
+   for (i = 0; i < count; i++) {
+      ((uint16 *)dst)[i] = val;
+   }
 #endif
    return dst;
 }
 
 
 /*
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  *
  * uint32set --
  *
@@ -533,16 +519,18 @@ uint16set(void *dst, uint16 val, size_t count)
  * Side effects:
  *      As with memset.
  *
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  */
 
-static INLINE void *
-uint32set(void *dst, uint32 val, size_t count)
+static inline void *
+uint32set(void *dst,    // OUT
+          uint32 val,   // IN
+          size_t count) // IN
 {
-#ifdef VM_ARM_32
+#if defined __GNUC__ && defined VM_ARM_32
    void *tmpDst = dst;
 
-   __asm__ __volatile__ (
+   __asm__ __volatile__(
       "cmp     %1, #0\n\t"
       "beq     2f\n"
       "1:\n\t"
@@ -553,14 +541,15 @@ uint32set(void *dst, uint32 val, size_t count)
       : "+r" (tmpDst), "+r" (count)
       : "r" (val)
       : "cc", "memory");
-#elif defined(VM_ARM_64)
+#elif defined __GNUC__ && defined VM_ARM_64
    void   *tmpDst = dst;
+   uint64  tmpVal = val;
 
    if (count == 0) {
       return dst;
    }
 
-   __asm__ __volatile__ (
+   __asm__ __volatile__(
       "cbz     %2, 1f\n\t"
 
       // Drop our value in the top 32 bits, then copy from there to the bottom
@@ -591,87 +580,31 @@ uint32set(void *dst, uint32 val, size_t count)
       "cbz     %1, 4f\n\t"
       "str     %w2, [%0]\n\t" // No incr
       "4:"
-      : "+r" (tmpDst), "+r" (count), "+r" (val)
+      : "+r" (tmpDst), "+r" (count), "+r" (tmpVal)
       :
       : "cc", "memory");
-#else
+#elif defined __GNUC__ && defined VM_X86_ANY
    size_t dummy0;
    void *dummy1;
 
-   __asm__ __volatile__("\t"
-                        "cld"            "\n\t"
-                        "rep ; stosl"    "\n"
-                        : "=c" (dummy0), "=D" (dummy1)
-                        : "0" (count), "1" (dst), "a" (val)
-                        : "memory", "cc"
-      );
-#endif
-   return dst;
-}
-
-#else /* unknown system: rely on C to write */
-static INLINE void *
-uint16set(void *dst, uint16 val, size_t count)
-{
+   __asm__ __volatile__(
+      "cld"            "\n\t"
+      "rep ; stosl"    "\n"
+      : "=c" (dummy0), "=D" (dummy1)
+      : "0" (count), "1" (dst), "a" (val)
+      : "memory", "cc");
+#elif defined _MSC_VER && defined VM_X86_ANY
+   __stosd((unsigned long *)dst, (unsigned long)val, count);
+#else /* Fall back to a plain C implementation. */
    size_t i;
+
    for (i = 0; i < count; i++) {
-     ((uint16 *) dst)[i] = val;
-   }
-   return dst;
-}
-
-static INLINE void *
-uint32set(void *dst, uint32 val, size_t count)
-{
-   size_t i;
-   for (i = 0; i < count; i++) {
-     ((uint32 *) dst)[i] = val;
-   }
-   return dst;
-}
-#endif // defined(VM_X86_ANY) || defined(VM_ARM_ANY)
-#elif defined(_MSC_VER)
-
-static INLINE void *
-uint16set(void *dst, uint16 val, size_t count)
-{
-#ifdef VM_X86_64
-   __stosw((uint16*)dst, val, count);
-#else
-   __asm { pushf;
-           mov ax, val;
-           mov ecx, count;
-           mov edi, dst;
-           cld;
-           rep stosw;
-           popf;
+      ((uint32 *)dst)[i] = val;
    }
 #endif
    return dst;
 }
 
-static INLINE void *
-uint32set(void *dst, uint32 val, size_t count)
-{
-#ifdef VM_X86_64
-   __stosd((unsigned long*)dst, (unsigned long)val, count);
-#else
-   __asm { pushf;
-           mov eax, val;
-           mov ecx, count;
-           mov edi, dst;
-           cld;
-           rep stosd;
-           popf;
-   }
-#endif
-   return dst;
-}
-
-#else
-#error "No compiler defined for uint*set"
-#endif
-
 
 /*
  *-----------------------------------------------------------------------------
@@ -686,8 +619,8 @@ uint32set(void *dst, uint32 val, size_t count)
 static INLINE uint16
 Bswap16(uint16 v)
 {
-#if defined(VM_ARM_64)
-   __asm__("rev16 %0, %0" : "+r"(v));
+#if defined(VM_ARM_64) && !defined(_MSC_VER)
+   __asm__("rev16 %w0, %w0" : "+r"(v));
    return v;
 #else
    return ((v >> 8) & 0x00ff) | ((v << 8) & 0xff00);
@@ -716,12 +649,12 @@ Bswap32(uint32 v) // IN
       : "0" (v)
    );
    return v;
-#elif defined(VM_ARM_32) && !defined(__ANDROID__)
+#elif defined(VM_ARM_32) && !defined(__ANDROID__) && !defined(_MSC_VER)
     __asm__("rev %0, %0" : "+r"(v));
     return v;
-#elif defined(VM_ARM_64)
-   __asm__("rev32 %0, %0" : "+r"(v));
-    return v;
+#elif defined(VM_ARM_64) && !defined(_MSC_VER)
+   __asm__("rev32 %x0, %x0" : "+r"(v));
+   return v;
 #else
    return    (v >> 24)
           | ((v >>  8) & 0xFF00)
@@ -745,72 +678,24 @@ Bswap32(uint32 v) // IN
 static INLINE uint64
 Bswap64(uint64 v) // IN
 {
-#if defined(VM_ARM_64)
+#if defined _MSC_VER
+   return _byteswap_uint64(v);
+#elif defined __GNUC__
+
+/* TODO: Return __builtin_bswap64(v) if gcc-arm64 is verified to use "rev". */
+#if defined VM_ARM_64
    __asm__("rev %0, %0" : "+r"(v));
    return v;
 #else
-   return ((uint64)Bswap((uint32)v) << 32) | Bswap((uint32)(v >> 32));
+   return __builtin_bswap64(v);
+#endif
+
+#else
+   return ((uint64)Bswap32((uint32)v) << 32) | Bswap32((uint32)(v >> 32));
 #endif
 }
 
 
-/*
- * COMPILER_MEM_BARRIER prevents the compiler from re-ordering memory
- * references accross the barrier.  NOTE: It does not generate any
- * instruction, so the CPU is free to do whatever it wants to...
- */
-#ifdef __GNUC__
-#define COMPILER_MEM_BARRIER()   __asm__ __volatile__ ("": : :"memory")
-#define COMPILER_READ_BARRIER()  COMPILER_MEM_BARRIER()
-#define COMPILER_WRITE_BARRIER() COMPILER_MEM_BARRIER()
-#elif defined(_MSC_VER)
-#define COMPILER_MEM_BARRIER()   _ReadWriteBarrier()
-#define COMPILER_READ_BARRIER()  _ReadBarrier()
-#define COMPILER_WRITE_BARRIER() _WriteBarrier()
-#endif
-
-
-/*
- *----------------------------------------------------------------------
- *
- * COMPILER_FORCED_LOAD_AND_MEM_BARRIER --
- *
- *        This macro prevents the compiler from re-ordering memory references
- *        across the barrier. In addition it emits a forced load from the given
- *        memory reference. The memory reference has to be either 1, 2, 4 or 8
- *        bytes wide.
- *        The forced load of a memory reference can be used exploit details of a
- *        given CPUs memory model. For example x86 CPUs won't reorder stores to
- *        a memory location x with loads from a memory location x.
- *        NOTE: It does not generate any fencing instruction, so the CPU is free
- *              to reorder instructions according to its memory model.
- *
- * Results:
- *        None
- *
- * Side Effects:
- *        None.
- *
- *----------------------------------------------------------------------
- */
-
-#ifdef VM_X86_64
-#ifdef __GNUC__
-
-#define COMPILER_FORCED_LOAD_AND_MEM_BARRIER(_memory_reference)               \
-   do {                                                                       \
-      typeof(_memory_reference) _dummy;                                       \
-                                                                              \
-      asm volatile("mov %1, %0\n\t"                                           \
-                   : "=r" (_dummy) /* Let compiler choose reg for _dummy */   \
-                   : "m" (_memory_reference)                                  \
-                   : "memory");                                               \
-   } while(0)
-
-#endif /* __GNUC__ */
-#endif /* VM_X86_64 */
-
-
 /*
  * PAUSE is a P4 instruction that improves spinlock power+performance;
  * on non-P4 IA32 systems, the encoding is interpreted as a REPZ-NOP.
@@ -818,29 +703,27 @@ Bswap64(uint64 v) // IN
  */
 static INLINE void
 PAUSE(void)
-#ifdef __GNUC__
+#if defined(__GNUC__) || defined(VM_ARM_32)
 {
-#ifdef VM_ARM_ANY
+#ifdef VM_ARM_64
+   __asm__ __volatile__("yield");
+#elif defined VM_ARM_32
    /*
-    * ARM has no instruction to execute "spin-wait loop", just leave it
-    * empty.
+    * YIELD is available in ARMv6K and above, so we could probably refine this
+    * instead of leaving it empty.
     */
-#else
+#else // x86
    __asm__ __volatile__( "pause" :);
 #endif
 }
 #elif defined(_MSC_VER)
-#ifdef VM_X86_64
 {
+#ifdef VM_X86_ANY
    _mm_pause();
+#else
+   __yield();
+#endif
 }
-#else /* VM_X86_64 */
-#pragma warning( disable : 4035)
-{
-   __asm _emit 0xf3 __asm _emit 0x90
-}
-#pragma warning (default: 4035)
-#endif /* VM_X86_64 */
 #else  /* __GNUC__  */
 #error No compiler defined for PAUSE
 #endif
@@ -875,7 +758,12 @@ RDTSC(void)
 
    return tim;
 #elif defined(VM_ARM_64)
-#if (defined(VMKERNEL) || defined(VMM)) && !defined(VMK_ARM_EL1)
+   /*
+    * Keep this implementation in sync with:
+    * bora/lib/vprobe/arm64/vp_emit_tc.c::VpEmit_BuiltinRDTSCWork()
+    * bora/modules/vmkernel/tests/core/xmapTest/xmapTest_arm64.c::XMapTest_SetupLoopCode()
+    */
+#if defined(VMKERNEL) && !defined(VMK_ARM_EL1_OR_VHE)
    return MRS(CNTPCT_EL0);
 #else
    return MRS(CNTVCT_EL0);
@@ -888,17 +776,20 @@ RDTSC(void)
 #endif
 }
 #elif defined(_MSC_VER)
-#ifdef VM_X86_64
+#ifdef VM_X86_ANY
 {
    return __rdtsc();
 }
 #else
-#pragma warning( disable : 4035)
 {
-   __asm _emit 0x0f __asm _emit 0x31
+   /*
+    * We need to do more inverstagetion here to find
+    * a microsoft equivalent of that code
+    */
+   NOT_IMPLEMENTED();
+   return 0;
 }
-#pragma warning (default: 4035)
-#endif /* VM_X86_64 */
+#endif /* VM_X86_ANY */
 #else  /* __GNUC__  */
 #error No compiler defined for RDTSC
 #endif /* __GNUC__  */
@@ -929,9 +820,10 @@ RDTSC(void)
 /*
  *-----------------------------------------------------------------------------
  *
- * {Clear,Set,Test}Bit{32,64} --
+ * {Clear, Set, Test, Toggle}Bit{32, 64} --
  *
- *    Sets or clears a specified single bit in the provided variable.
+ *    Sets tests clears or toggles a specified single bit in the provided
+ *    variable.
  *
  *    The index input value specifies which bit to modify and is 0-based.
  *    Index is truncated by hardware to a 5-bit or 6-bit offset for the
@@ -945,111 +837,51 @@ RDTSC(void)
  */
 
 static INLINE void
-SetBit32(uint32 *var, uint32 index)
+SetBit32(uint32 *var, unsigned index)
 {
-#if defined(__GNUC__) && defined(VM_X86_ANY)
-   __asm__ (
-      "bts %1, %0"
-      : "+mr" (*var)
-      : "rI" (index)
-      : "cc"
-   );
-#elif defined(_MSC_VER)
-   _bittestandset((long *)var, index);
-#else
-   *var |= (1 << index);
-#endif
+   *var |= 1 << index;
 }
 
 static INLINE void
-ClearBit32(uint32 *var, uint32 index)
+ClearBit32(uint32 *var, unsigned index)
 {
-#if defined(__GNUC__) && defined(VM_X86_ANY)
-   __asm__ (
-      "btr %1, %0"
-      : "+mr" (*var)
-      : "rI" (index)
-      : "cc"
-   );
-#elif defined(_MSC_VER)
-   _bittestandreset((long *)var, index);
-#else
    *var &= ~(1 << index);
-#endif
 }
 
 static INLINE void
-SetBit64(uint64 *var, uint64 index)
+ToggleBit32(uint32 *var, unsigned index)
 {
-#if defined(VM_64BIT) && !defined(VM_ARM_64)
-#ifdef __GNUC__
-   __asm__ (
-      "bts %1, %0"
-      : "+mr" (*var)
-      : "rJ" (index)
-      : "cc"
-   );
-#elif defined(_MSC_VER)
-   _bittestandset64((__int64 *)var, index);
-#endif
-#else
-   *var |= ((uint64)1 << index);
-#endif
+   *var ^= 1 << index;
 }
 
 static INLINE void
-ClearBit64(uint64 *var, uint64 index)
+SetBit64(uint64 *var, unsigned index)
 {
-#if defined(VM_64BIT) && !defined(VM_ARM_64)
-#ifdef __GNUC__
-   __asm__ (
-      "btrq %1, %0"
-      : "+mr" (*var)
-      : "rJ" (index)
-      : "cc"
-   );
-#elif defined(_MSC_VER)
-   _bittestandreset64((__int64 *)var, index);
-#endif
-#else
-   *var &= ~((uint64)1 << index);
-#endif
+   *var |= CONST64U(1) << index;
+}
+
+static INLINE void
+ClearBit64(uint64 *var, unsigned index)
+{
+   *var &= ~(CONST64U(1) << index);
+}
+
+static INLINE void
+ToggleBit64(uint64 *var, unsigned index)
+{
+   *var ^= (CONST64U(1) << index);
 }
 
 static INLINE Bool
-TestBit32(const uint32 *var, uint32 index)
+TestBit32(const uint32 *var, unsigned index)
 {
-#if defined(__GNUC__) && defined(VM_X86_ANY)
-   Bool bit;
-   __asm__ (
-      "bt %[index], %[var] \n"
-      "setc %[bit]"
-      : [bit] "=qQm" (bit)
-      : [index] "rI" (index), [var] "r" (*var)
-      : "cc"
-   );
-   return bit;
-#else
    return (*var & (1 << index)) != 0;
-#endif
 }
 
 static INLINE Bool
-TestBit64(const uint64 *var, uint64 index)
+TestBit64(const uint64 *var, unsigned index)
 {
-#if defined __GNUC__ && defined VM_X86_64
-   Bool bit;
-   __asm__ (
-      "bt %[index], %[var] \n"
-      "setc %[bit]"
-      : [bit] "=qQm" (bit)
-      : [index] "rJ" (index), [var] "r" (*var)
-      : "cc"
-   );
-   return bit;
-#else
    return (*var & (CONST64U(1) << index)) != 0;
-#endif
 }
 
 /*
@@ -1067,18 +899,27 @@ TestBit64(const uint64 *var, uint64 index)
  *-----------------------------------------------------------------------------
  */
 
+#if defined __GCC_ASM_FLAG_OUTPUTS__
+/*
+ * See https://gcc.gnu.org/onlinedocs/gcc/Extended-Asm.html
+ * 6.47.2.4 Flag Output Operands
+ *
+ * This expands to 0 or 1 instructions followed by the output operand string.
+ */
+#define GCC_ASM_BT_EPILOG : "=@ccc"
+#else
+#define GCC_ASM_BT_EPILOG "\n\tsetc\t%0" : "=qQm"
+#endif
+
 static INLINE Bool
 SetBitVector(void *var, int32 index)
 {
 #if defined(__GNUC__) && defined(VM_X86_ANY)
    Bool bit;
-   __asm__ (
-      "bts %2, %1;"
-      "setc %0"
-      : "=qQm" (bit), "+m" (*(uint32 *)var)
-      : "rI" (index)
-      : "memory", "cc"
-   );
+   __asm__("bts\t%2, %1"
+           GCC_ASM_BT_EPILOG (bit), "+m" (*(uint32 *)var)
+           : "rI" (index)
+           : "memory", "cc");
    return bit;
 #elif defined(_MSC_VER)
    return _bittestandset((long *)var, index) != 0;
@@ -1094,13 +935,10 @@ ClearBitVector(void *var, int32 index)
 {
 #if defined(__GNUC__) && defined(VM_X86_ANY)
    Bool bit;
-   __asm__ (
-      "btr %2, %1;"
-      "setc %0"
-      : "=qQm" (bit), "+m" (*(uint32 *)var)
-      : "rI" (index)
-      : "cc"
-   );
+   __asm__("btr\t%2, %1"
+           GCC_ASM_BT_EPILOG (bit), "+m" (*(uint32 *)var)
+           : "rI" (index)
+           : "memory", "cc");
    return bit;
 #elif defined(_MSC_VER)
    return _bittestandreset((long *)var, index) != 0;
@@ -1116,13 +954,10 @@ ComplementBitVector(void *var, int32 index)
 {
 #if defined(__GNUC__) && defined(VM_X86_ANY)
    Bool bit;
-   __asm__ (
-      "btc %2, %1;"
-      "setc %0"
-      : "=qQm" (bit), "+m" (*(uint32 *)var)
-      : "rI" (index)
-      : "cc"
-   );
+   __asm__("btc\t%2, %1"
+           GCC_ASM_BT_EPILOG (bit), "+m" (*(uint32 *)var)
+           : "rI" (index)
+           : "memory", "cc");
    return bit;
 #elif defined(_MSC_VER)
    return _bittestandcomplement((long *)var, index) != 0;
@@ -1138,13 +973,10 @@ TestBitVector(const void *var, int32 index)
 {
 #if defined(__GNUC__) && defined(VM_X86_ANY)
    Bool bit;
-   __asm__ (
-      "bt %2, %1;"
-      "setc %0"
-      : "=qQm" (bit)
-      : "m" (*(const uint32 *)var), "rI" (index)
-      : "cc"
-   );
+   __asm__("bt\t%2, %1"
+           GCC_ASM_BT_EPILOG (bit)
+           : "m" (*(const uint32 *)var), "rI" (index)
+           : "cc");
    return bit;
 #elif defined _MSC_VER
    return _bittest((long *)var, index) != 0;
@@ -1153,6 +985,45 @@ TestBitVector(const void *var, int32 index)
 #endif
 }
 
+#undef GCC_ASM_BT_EPILOG
+
+/*
+ *-----------------------------------------------------------------------------
+ * RoundDownPow2_{64,32} --
+ *
+ *   Rounds a value down to the previous power of 2.  Returns the original
+ *   value if it is a power of 2. Returns 0 for input of 0 and 1 for 1.
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE uint64
+RoundDownPow2_64(uint64 value)
+{
+   if ((value & (value - 1)) == 0) {
+      /*
+       * Already zero or a power of two.
+       */
+      return value;
+   }
+
+   return CONST64U(1) << mssb64_0(value);
+}
+
+
+static INLINE uint32
+RoundDownPow2_32(uint32 value)
+{
+   if ((value & (value - 1)) == 0) {
+      /*
+       * Already a power of two.
+       */
+      return value;
+   }
+
+   return 1U << mssb32_0(value);
+}
+
+
 /*
  *-----------------------------------------------------------------------------
  * RoundUpPow2_{64,32} --
@@ -1213,7 +1084,9 @@ RoundUpPow2C32(uint32 value)
    if (value <= 1 || value > (1U << 31)) {
       return 1; // Match the assembly's undefined value for large inputs.
    } else {
-      return (2 << mssb32_0(value - 1));
+      int mssb32 = mssb32_0(value - 1);
+      /* invariant: mssb32 >= 0 */
+      return (2U << (uint32)mssb32);
    }
 }
 
@@ -1234,7 +1107,7 @@ RoundUpPow2Asm32(uint32 value)
                                            // if out == 2^32 then out = 1 as it is right rotate
        : [in]"+r"(value),[out]"+r"(out));
    return out;
-#elif defined(VM_ARM_64)
+#elif defined(VM_ARM_64) || defined(__wasm__)
    return RoundUpPow2C32(value);
 #else
    uint32 out = 2;
@@ -1268,5 +1141,190 @@ RoundUpPow2_32(uint32 value)
 #endif
 }
 
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * PopCount32 --
+ *
+ *     Counts "1" bits in a uint32.
+ *
+ * Results:
+ *     Returns the number of bits set to 1.
+ *
+ * Side effects:
+ *     None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE unsigned
+PopCount32(uint32 value)
+{
+#if defined(__GNUC__) && defined(__POPCNT__)
+   return __builtin_popcount(value);
+#else
+   /*
+    * Attribution:
+    *     This algorithm was copied from:
+    *         http://www.aggregate.org/MAGIC#Population Count (Ones Count)
+    *
+    *     A virtually identical version (but in assembly) appears in an
+    *     AMD reference manual.
+    *
+    *     No license appears in the original code, but the website
+    *     header states:
+    *
+    *     "None of the following coding tricks came from proprietary
+    *     sources; further, we believe that each of the tricks we did
+    *     not invent is essentially "standard engineering practice" in
+    *     the specialized niche where it applies. Thus, although we
+    *     have not conducted patent searches, etc., to confirm it, we
+    *     believe that these are tricks that freely can be used for
+    *     any purpose. Of course, The Aggregate accepts no
+    *     responsibility for your use of these tricks; you must
+    *     confirm that the trick does what you want and that you can
+    *     use it as you intend. That said, we do intend to maintain
+    *     this page by adding new algorithms and/or correcting
+    *     existing entries. If you have any comments, please contact
+    *     Professor Hank Dietz, http://aggregate.org/hankd/"
+    *
+    *     "This document should be cited using something like the
+    *     following bibtex entry:" (most recent retrieval date added)
+    *
+    *     @techreport{magicalgorithms,
+    *     author={Henry Gordon Dietz},
+    *     title={{The Aggregate Magic Algorithms}},
+    *     institution={University of Kentucky},
+    *     howpublished={Aggregate.Org online technical report},
+    *     URL={http://aggregate.org/MAGIC/},
+    *     urldate={2016-01-27}
+    *     }
+    */
+   value -= ((value >> 1) & 0x55555555);
+   value = (((value >> 2) & 0x33333333) + (value & 0x33333333));
+   value = (((value >> 4) + value) & 0x0f0f0f0f);
+   value += (value >> 8);
+   value += (value >> 16);
+   return value & 0x0000003f;
+#endif
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * PopCount64 --
+ *
+ *     Counts "1" bits in a uint64.
+ *
+ * Results:
+ *     Returns the number of bits set to 1.
+ *
+ * Side effects:
+ *     None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE unsigned
+PopCount64(uint64 value)
+{
+#if defined(__GNUC__) && defined(__POPCNT__)
+#if defined(VM_X86_64)
+   return __builtin_popcountll(value);
+#else
+   return PopCount32(value) + PopCount32(value >> 32);
+#endif
+#else
+   value -= (value >> 1) & 0x5555555555555555ULL;
+   value = ((value >> 2) & 0x3333333333333333ULL) +
+           (value & 0x3333333333333333ULL);
+   value = ((value >> 4) + value) & 0x0f0f0f0f0f0f0f0fULL;
+   value += value >> 8;
+   value += value >> 16;
+   value += value >> 32;
+   return (unsigned) (value & 0xff);
+#endif
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * INTR_R_BARRIER_R --
+ * INTR_R_BARRIER_W --
+ * INTR_R_BARRIER_RW --
+ * INTR_W_BARRIER_R --
+ * INTR_W_BARRIER_W --
+ * INTR_W_BARRIER_RW --
+ * INTR_RW_BARRIER_R --
+ * INTR_RW_BARRIER_W --
+ * INTR_RW_BARRIER_RW --
+ *
+ *      Enforce ordering on memory operations witnessed by and
+ *      affected by interrupt handlers.
+ *
+ *      This should be used to replace the legacy COMPILER_MEM_BARRIER
+ *      for code that has been audited to determine it only needs
+ *      ordering with respect to interrupt handlers, and not to other
+ *      CPUs (SMP_*), memory-mapped I/O (MMIO_*), or DMA (DMA_*).
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+#ifdef __GNUC__
+
+static INLINE void
+INTR_RW_BARRIER_RW(void)
+{
+   __asm__ __volatile__("" ::: "memory");
+}
+
+#define INTR_R_BARRIER_R INTR_RW_BARRIER_RW
+#define INTR_R_BARRIER_W INTR_RW_BARRIER_RW
+#define INTR_R_BARRIER_RW INTR_RW_BARRIER_RW
+#define INTR_W_BARRIER_R INTR_RW_BARRIER_RW
+#define INTR_W_BARRIER_W INTR_RW_BARRIER_RW
+#define INTR_W_BARRIER_RW INTR_RW_BARRIER_RW
+#define INTR_RW_BARRIER_R INTR_RW_BARRIER_RW
+#define INTR_RW_BARRIER_W INTR_RW_BARRIER_RW
+
+#elif defined _MSC_VER
+
+static INLINE void
+INTR_R_BARRIER_R(void)
+{
+   _ReadBarrier();
+}
+
+static INLINE void
+INTR_W_BARRIER_W(void)
+{
+   _WriteBarrier();
+}
+
+static INLINE void
+INTR_RW_BARRIER_RW(void)
+{
+   _ReadWriteBarrier();
+}
+
+#define INTR_R_BARRIER_W INTR_RW_BARRIER_RW
+#define INTR_R_BARRIER_RW INTR_RW_BARRIER_RW
+#define INTR_W_BARRIER_R INTR_RW_BARRIER_RW
+#define INTR_W_BARRIER_RW INTR_RW_BARRIER_RW
+#define INTR_RW_BARRIER_R INTR_RW_BARRIER_RW
+#define INTR_RW_BARRIER_W INTR_RW_BARRIER_RW
+
+#else
+#error No compiler defined for INTR_*_BARRIER_*
+#endif
+
+
+#if defined __cplusplus
+} // extern "C"
+#endif
+
 #endif // _VM_BASIC_ASM_H_
 
